/*
The copyright in this software is being made available under this MPEG Reference Software Copyright License. This software may be subject to other third party and contributor rights, including patent rights, and no such rights are granted under this license.

Copyright (c) 2016, Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
*	Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
*	Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
*	Neither the names of Joanneum Research, Mitsubishi Electric Research Labs, Peking University, Telecom Italia, University of Surrey, Visual Atoms nor the names of their contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include "CdvaImpl.h"

#include <iostream>
#include <sstream>
#include <fstream>
#include <opencv2/opencv.hpp>
#include <algorithm>

using namespace cv;
using namespace mpeg7cdva;
using namespace mpeg7cdvs;
using namespace std;

#define FOURCC_MP43 0x3334504D		// identify MS codec MP43

///////////////////////////////// class MCdvsDescriptor //////////////////////////////////

void MCdvsDescriptor::setStartTimeMs(unsigned long position_ms)
{
	startTime = position_ms;
}

void MCdvsDescriptor::setEndTimeMs(unsigned long position_ms)
{
	endTime = position_ms;
}


unsigned long MCdvsDescriptor::getStartTimeMs() const
{
	return startTime;
}

unsigned long MCdvsDescriptor::getEndTimeMs() const
{
	return endTime;
}

size_t MCdvsDescriptor::write(std::ofstream & fout) const
{
	fout.write((char*)&startTime, sizeof(long));
	fout.write((char*)&endTime, sizeof(long));

	// write buffer size
	unsigned long mysize = (unsigned long) buffer.size();
	fout.write((char*)&mysize, sizeof(long));

	// write descriptor
	fout.write((char*)buffer.data(), buffer.size());

	return (3*sizeof(long) + mysize);
}

size_t MCdvsDescriptor::read(std::ifstream & fin)
{
	fin.read((char*)&startTime, sizeof(long));
	fin.read((char*)&endTime, sizeof(long));

	unsigned long mysize;
	fin.read((char*)&mysize, sizeof(long));

	if (fin.good())
	{
		buffer.resize(mysize);
		fin.read((char*)buffer.data(), mysize);
		return (3*sizeof(long) + mysize);
	}
	else
		return 0;		// end of file
}

///////////////////////////////// class CdvaImpl //////////////////////////////////


CdvaImpl::CdvaImpl():verboseMode(false),current_op(UNKNOWN),skip_before(4), skip_after(4), drop_frame_th(20), max_retrieved(50)
{
	cdvsconfig = NULL;
	cdvsclient = NULL;
	cdvsserver = NULL;
}

CdvaImpl::~CdvaImpl()
{
	delete cdvsclient;
	delete cdvsserver;
	delete cdvsconfig;
}

void CdvaImpl::init(OPERATION op, bool verbose, size_t n_videos, int querybitrate, int refbitrate)
{
	current_op = op;
	verboseMode = verbose;

	int cdvsMode = 0;		// cdvsMode = 6 is also optionally allowed

	cdvsconfig = CdvsConfiguration::cdvsConfigurationFactory();

	if (current_op == EXTRACT)
		cdvsclient = CdvsClient::cdvsClientFactory(cdvsconfig, cdvsMode); // use ALP

	cdvsserver = CdvsServer::cdvsServerFactory(cdvsconfig);		// used in extract, match and retrieve

	switch (querybitrate) {
	case  0:
	case 16:
		skip_before = 4;
		skip_after = 4;
		drop_frame_th = 21;
		max_retrieved = 50;
		break;
	case 64:
		skip_before = 3;
		skip_after = 3;
		drop_frame_th = 28;
		max_retrieved = 50;
		break;
	case 256:
		skip_before = 2;
		skip_after = 2;
		drop_frame_th = 36;
		max_retrieved = 50;
		break;
	}

	// prepare DB (in case retrieval is performed)
	if (current_op == RETRIEVE)
		cdvsserver->createDB(cdvsMode, n_videos);
}

void CdvaImpl::close()
{
	if (current_op == RETRIEVE)
		cdvsserver->clearDB();
}

/*
* Check valid bitrates
*/
bool CdvaImpl::checkBitrate(int bitrate)
{
	switch (bitrate)
	{
	case   0: return true;
	case  16: return true;
	case  64: return true;
	case 256: return true;
	}
	return false;
}

inline int skipFrames(int framenum, VideoCapture & vc)
{
	int skipped = 0;
	while ((skipped < framenum) && vc.grab())
		skipped++;

	return skipped;
}

inline double getCurrentTimeMillis(VideoCapture & vc)
{
	return ((int) vc.get(CV_CAP_PROP_FOURCC) == FOURCC_MP43)? 100 * vc.get(CV_CAP_PROP_POS_MSEC) : vc.get(CV_CAP_PROP_POS_MSEC);	// fix codec-specific issues
}

inline double getCurrentTimeSeconds(VideoCapture & vc)
{
	return ((int) vc.get(CV_CAP_PROP_FOURCC) == FOURCC_MP43)? 0.1 * vc.get(CV_CAP_PROP_POS_MSEC) : 0.001 * vc.get(CV_CAP_PROP_POS_MSEC);	// fix codec-specific issues
}


/*
* CDVA descriptor extension; may be changed (if needed) to identify a specific implementation.
*/
const char * CdvaImpl::getDescriptorExt(int bitrate)
{
	switch (bitrate)
	{
	case   0: return ".DB.cdva";
	case  16: return ".16K.cdva";
	case  64: return ".64K.cdva";
	case 256: return ".256K.cdva";
	}
	return NULL;
}

string CdvaImpl::getExt(const string & imageName)
{
	size_t found = imageName.find_last_of('.');
	return imageName.substr(found + 1, imageName.length());
}

/*
* CDVA extract implementation
*/
void CdvaImpl::extract(const std::string & descrname, const std::string & videopathname, int mode, ExtractData & outdata) const
{
	std::ofstream fout;
	fout.open(descrname.c_str(), std::ofstream::binary);
	if (! fout.is_open())
	{
		throw CdvaException(string("Cannot write: ").append(descrname));
	}

	// check the filename extension to discriminate images from videos

	string ext = getExt(videopathname);
	if (ext.compare("jpeg") == 0||ext.compare("jpg") == 0||ext.compare("bmp") == 0||ext.compare("png") == 0||ext.compare("tiff") == 0||ext.compare("tif") == 0)
	{
		Mat greyimage = imread(videopathname, CV_LOAD_IMAGE_GRAYSCALE);
		MCdvsDescriptor output;
		cdvsclient->encode(output, greyimage.cols, greyimage.rows, greyimage.data);
		output.setStartTimeMs(0);
		output.setEndTimeMs(1000);
		output.write(fout);

		outdata.setNumFrames(1);					// 1 picture
		outdata.setVideoDuration(1.0);				// 1 second (conventional value for a picture)

		if (verboseMode)
		{
			cout << "processed " << videopathname << " image " << endl;
		}

	}
	else	// assuming it is a video clip
	{
		VideoCapture videocpp(videopathname);

		if (!videocpp.isOpened())
		{
			throw CdvaException(string("File not found: ").append(videopathname));
		}

		double tot_frames = 0;		// do not trust the video header information: compute the number of frames that can be read

		// MAIN LOOP: encode all frames using CDVA

		Mat colorimage;				// input image
		Mat greyimage;				// converted greyscale image

		MCdvsDescriptor prevDesc;

		int c1 = 0;
		int c2 = 0;
		int current_frame = 0;
		bool initial_frame = true;

		do
		{
			MCdvsDescriptor desc;

			desc.setStartTimeMs((unsigned long) getCurrentTimeMillis(videocpp));

			tot_frames += (c1 = skipFrames(skip_before, videocpp));		// skip n frames

			if (!videocpp.read(colorimage)) break;		// read one frame

			current_frame = ++tot_frames; 			// must be done before setting c2!

			tot_frames += (c2 = skipFrames(skip_after, videocpp));		// skip n frames

			desc.setEndTimeMs((unsigned long) getCurrentTimeMillis(videocpp));

			cvtColor(colorimage, greyimage, cv::COLOR_BGR2GRAY);		// convert to luminance only

			cdvsclient->encode(desc, greyimage.cols, greyimage.rows, greyimage.data);

			bool drop_frame = false;

			if (initial_frame)
			{
				desc.write(fout);			// write the first encoded descriptor
				initial_frame = false;		// next frames are not initial
				prevDesc = desc;			// save descriptor instance
			}
			else
			{
				PointPairs pp = cdvsserver->match(desc, prevDesc, NULL, NULL, MATCH_TYPE_GLOBAL);	// match using Global Descriptor
				if (pp.global_score > drop_frame_th)
				{
					drop_frame = true;
				}
				else
				{
					desc.write(fout);			// write encoded descriptor
					prevDesc = desc;			// save descriptor instance
				}
			}

			if (verboseMode)
			{
				cout << (drop_frame? "dropped frame " : "processed frame ") << current_frame
					 << " start ms: " << desc.getStartTimeMs() << " end ms: " << desc.getEndTimeMs() << endl;
			}

		}	while ((c1 == skip_before) && (c2 == skip_after));		// stop if video does not contain other frames

		double time_s = getCurrentTimeSeconds(videocpp);	// normalized time in seconds
		outdata.setVideoDuration(time_s);
		outdata.setNumFrames(tot_frames);
	}	// end else

	outdata.setDescriptorLength((double) fout.tellp());		// real descriptor size (in bytes)
	fout.close();
}

/*
* CDVA match implementation:
*/
double CdvaImpl::match(MatchData & matchResults, const std::string & qdescrname, const std::string & rdescrname, int qmode, int rmode)
{
	if (verboseMode) {
		cout << "matching " << qdescrname << " against " << rdescrname << endl;
	}

	// pairwise matching of frame descriptors
	MCdvsDescriptorList qDescList, rDescList;

	qDescList.reserve(100);
	rDescList.reserve(100);

	parse(qdescrname, qDescList);
	parse(rdescrname, rDescList);

	return match(matchResults, qDescList, rDescList);


}


double CdvaImpl::match(MatchData & matchResults, const MCdvsDescriptorList& qDescList, const MCdvsDescriptorList& rDescList)
{
	double maxScore = 0;

	for (int i = 0; i < qDescList.size(); i ++) {

		for (int j = 0; j < rDescList.size(); j ++) {

			PointPairs ppglobal = cdvsserver->match(qDescList[i], rDescList[j], NULL, NULL, MATCH_TYPE_GLOBAL);	// match using Global Descriptor

			if (ppglobal.global_score > ppglobal.global_threshold)
			{
				PointPairs pplocal = cdvsserver->match(qDescList[i], rDescList[j], NULL, NULL, MATCH_TYPE_LOCAL); // match using Local Descriptors
				if (pplocal.local_score > pplocal.local_threshold)
				{
					// there is a match
					double score = ppglobal.global_score * pplocal.local_score;

					if (score > maxScore)
						maxScore = score;

					matchResults.setMatchingTime(0.001 * qDescList[i].getStartTimeMs());	// set the query temporal localization (start time)
					matchResults.setMatchingTime(0.001 * qDescList[i].getEndTimeMs());		// set the query temporal localization (end time)
				}
			}
		}
	}

	matchResults.setMatchingScore(maxScore);

	return maxScore;
}

/*
* CDVA makeindex implementation: can be changed freely;
* you can signal errors at any point of your code using "throw CdvaException(string);"
* the following default implementation of makeindex just appends id + descriptor data to the index file.
*/
void  CdvaImpl::makeindex(const std::string & cdva_descriptor, const std::string & relativepathname)
{
	// read descriptor and append it to the DB
	
	std::string descriptorId = relativepathname;
	MCdvsDescriptorList descriptors;
	parse(cdva_descriptor, descriptors); // read descriptors from file

	for (int i = 0; i < descriptors.size(); i++) {
		// ignore descriptors with no local features
		if (descriptors[i].featurelist.nFeatures() > 0) {
			int idx = cdvsserver->addDescriptorToDB(descriptors[i], descriptorId.c_str());	// add image to DB
		}
	}

	cdvsserver->commitDB();

}

//
// order MatchData elements by descending score
//
bool CdvaImpl::byDescendingScore(const MatchData & m1, const MatchData & m2) {
	return (m1.getScore() > m2.getScore());
}

/*
* CDVA retrieve implementation: can be changed freely;
* you can signal errors at any point of your code using "throw CdvaException(string);"
*/
void CdvaImpl::retrieve(vector<MatchData> & retrievalResults, const std::string & qdescrname, int qmode)
{
	MCdvsDescriptorList qDescList;

	parse(qdescrname, qDescList);

	retrievalResults.clear();

	map<string, MatchData> matches; // matches per video

	for (int i = 0; i < qDescList.size(); i++) {
		vector<RetrievalData> cdvsResults;
		int n = cdvsserver->retrieve(cdvsResults, qDescList[i], 100);

		// process results (this version saves the top match even if fScore == 0 to avoid empty results)
		for (unsigned int k = 0; k<n; ++k)
		{
			if ((k == 0) || (cdvsResults[k].fScore > 0)) {

				string imgname = cdvsserver->getImageId(cdvsResults[k].index);
				map<string, MatchData>::iterator iter = matches.find(imgname);
				if (iter == matches.end()) {
					MatchData md;
					md.setReferenceID(imgname);
					md.setMatchingScore(cdvsResults[k].fScore);
					md.setMatchingTime(0.001 * qDescList[i].getStartTimeMs());		// (optional) matching time is always related to the query (the reference is known)
					md.setMatchingTime(0.001 * qDescList[i].getEndTimeMs());		// (optional)
					matches[imgname] = md;
				}
				else {
					if (cdvsResults[k].fScore > iter->second.getScore()) iter->second.setMatchingScore(cdvsResults[k].fScore);
					iter->second.setMatchingTime(0.001 * qDescList[i].getStartTimeMs());	// optional
					iter->second.setMatchingTime(0.001 * qDescList[i].getEndTimeMs());		// optional
				}
			}
		}

	}

	// copy remaining match data entries
	for (map<string, MatchData>::iterator it = matches.begin(); it != matches.end(); ++it) {
		retrievalResults.push_back(it->second);
	}

	// check the number of retrieved images - cannot be more than max_retrieved

	if (retrievalResults.size() > max_retrieved)
	{
		// partially sort results by descending score
		partial_sort(retrievalResults.begin(), retrievalResults.begin() + max_retrieved, retrievalResults.end(), byDescendingScore);
		retrievalResults.resize(max_retrieved);		// limit results to the top 50
	}
	else	// size is less or equal than the threshold
	{
		// sort all results by descending score
		sort(retrievalResults.begin(), retrievalResults.end(), byDescendingScore);
	}


}


void CdvaImpl::parse(const std::string & descFile, MCdvsDescriptorList& descList) {

	ifstream reader(descFile.c_str(), std::ifstream::binary);

	if (!reader.good())
		throw CdvaException(string("Cannot read descriptor: ").append(descFile));

	MCdvsDescriptor desc;

	while (desc.read(reader) > 0) {
		cdvsserver->decode(desc);
		descList.push_back(desc);
	}

	reader.close();

}
